/*
**    FILE    : galois.h   
**    PURPOSE : Header-file to galois.c
**    AUTHOR  : Bart De Canne, Belgium (bdc@barco.be)
**    COMPILER: BC++ 3.1 compiler
**    DEPENDENCIES:
**    DATE    : 01-09-1994
**    VERSION : 1.0
*/


#ifndef _GALOIS_HEADER
#define _GALOIS_HEADER

#include <alloc.h>
#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <mem.h>
#include <math.h>
#include <dos.h>
#include <io.h>
#include <process.h>

/* -----------------------  macros --------------------------------------- */

#define uint unsigned int
#define ulong unsigned long
#define byte unsigned char
#define TRUE  1
#define FALSE 0
#define bool byte
#define MAX(x,y) ((x)>(y) ? (x) : (y))
#define MIN(x,y) ((x)<(y) ? (x) : (y))
#define MAXENTRIES 1024
#define STRING_LENGTH 30

/* 
** Since + & . operations commute,
** it is sufficient to generate only one half of the NxN-dimensional tables
** This reduces memory space. Proper indexing into the 
** 2-D tables is done using the following macros.
*/ 

#define I(i,j) ((i>j) ? j : i)
#define J(i,j) ((i>j) ? (i-j) : (j-i))
#define MULT(i,j) Mult[I(i,j)][J(i,j)]
#define ADD(i,j) Add[I(i,j)][J(i,j)]

/* 
**  affects the fuctionality of MultPoly() 
*/ 

#define COPY      1
#define SUM       2
#define SUBTRACT  3

/* 
** macros for error reporting 
*/ 

#define BEGIN(name) static char RoutineName[]= name
#define ERROR   {printf("ERROR: F>%s:R>%s:L>%d: ",__FILE__,                \
                  RoutineName,__LINE__);                                   \
                 exit(-1);                                                 \
                }

/* 
** gets degree of polynomial and stores it in <pp>->Deg
*/ 

#define DEGREE(pp)   {                                                     \
                     for(pp->Deg=pp->MaxDeg;pp->Deg>=0;pp->Deg--) {        \
                        if(pp->c[pp->Deg]) break;                          \
                        }                                                  \
                     if(pp->Deg<0) pp->Deg=0;                              \
                     }
/* 
** returns TRUE if <pp> is the all-zeros polynomial
*/ 

#define ISZERO(pp)   ((!pp->Deg && !pp->c[0]) ? TRUE : FALSE)

/* 
** sets polynomial <pp> equal to <arg> in degree <deg>
*/ 

#define SET_POLY(pp,deg,arg)  { memset(pp->c,0,sizeof(int)*(pp->MaxDeg+1));\
                                pp->c[deg]=arg;                            \
                                pp->Deg=deg;                               \
                              }

/* 
** macros for pacifier during calculations
*/ 

static byte toggle=FALSE;
static int old_x,old_y;


#define START_PACIFIER {   old_x=wherex();old_y=wherey();}

#define PACIFIER(perc) {                                                   \
   gotoxy(55,1);                                                           \
   if(!toggle) {                                                           \
      fputs("Calculating...",stdout);                                      \
      if(perc) printf("(%3d%%)",(int)perc);                                \
      else fputs("       ",stdout);                                        \
      toggle=TRUE;                                                         \
      }                                                                    \
   else {                                                                  \
      fputs("Wait please...",stdout);                                      \
      if(perc) printf("(%3d%%)",(int)perc);                                \
      else fputs("       ",stdout);                                        \
      toggle=FALSE;                                                        \
      }                                                                    \
   }

#define STOP_PACIFIER {                                                    \
   gotoxy(55,1);                                                           \
   puts("                    ");                                           \
   toggle=FALSE;                                                           \
   gotoxy(old_x,old_y);                                                    \
   }

/* 
** macros for reading/writing data-block
*/ 

#define READ_PP(pp) {                                                      \
                                                                           \
   static int n;                                                           \
                                                                           \
   n=pp->MaxDeg+1;                                                         \
   while(--n>=0)                                                           \
      pp->c[n]=GetElementIndex(GetBits(bps));                              \
   DEGREE(pp);                                                             \
   }

#define WRITE_PP(pp) {                                                     \
                                                                           \
   static int n,k;                                                         \
                                                                           \
   n=pp->MaxDeg+1;                                                         \
   while(--n>=0) {                                                         \
      k=bps;                                                               \
      while(--k>=0)                                                        \
         PutBits(pg_s->pe[pp->c[n]].Alpha[k],1);                           \
      }                                                                    \
   }


/* ------------------------  structures  --------------------------------- */ 

#define POLYNOM   struct tag_POLYNOM
#define POLYARR   struct tag_POLYARR
#define ELEMENT   struct tag_ELEMENT
#define GALOIS    struct tag_GALOIS
#define OPTABLE   struct tag_OPTABLE


/* Polynomial */

POLYNOM { 
         int MaxDeg;          /* max.degree of dyn.allocated polynomial    */
         int Deg;             /* max.degree of actual polynomial stored    */
         int Base;            /* p of GF(p^q) -> coeff. of polynom:0..p-1  */
         int *c;              /* array of coeff.values                     */
         };
typedef POLYNOM *pPOLYNOM;
typedef POLYNOM **ppPOLYNOM;


/* Array of polynomials */

POLYARR {
         int MaxNo;           /* max. no of polynomials in array           */
         int FirstFree;       /* entry of first free element in array      */
         ppPOLYNOM ppp;       /* array of pointers to polynom              */
         };               
typedef POLYARR *pPOLYARR;


/* Element of Galois Field and its minimal polynomial */

ELEMENT {                     /* denotes representation of field           */
         byte Alpha[8];       /* elements and their respective minimal     */
         pPOLYNOM pp;         /* polynomials                               */
         };
typedef ELEMENT *pELEMENT;


/* Galois Field   */

GALOIS {
         int Base;            /* p of GF(p^q)                              */
         int Exp;             /* q of GF(p^q)                              */
         int MaxNo;           /* p^q                                       */
         pELEMENT pe;         /* array of <GFElement>'s, of length         */
         };                   /* <MaxNo>                                   */
typedef GALOIS *pGALOIS;


/* Table defining operations in a Galois Field  */

OPTABLE   {                   /* Arrays defining operations in GF(p^q)     */
      int MaxNo;              
      int Base;
      int Exp;
      byte **Add;             /* Addition                                  */
      byte **Mult;            /* Multiplication                            */
      byte *InvAdd;           /* Inverses for addition                     */
      byte *InvMult;          /* Inverses for multiplication               */
      };
typedef OPTABLE *pOPTABLE;

/* ---------------------- function prototypes ---------------------------- */

#include "prototyp.h"


/* ---------------------- globals ---------------------------------------- */

EXT_MAINPROG pGALOIS  pg_s;   /* SYMBOL field GF(q)                        */
EXT_MAINPROG pGALOIS  pg_l;   /* LOCATOR field GF(q^n)                     */
EXT_MAINPROG pOPTABLE po_s;   /* operations in GF(q)                       */
EXT_MAINPROG pOPTABLE po_l;   /* operations in GF(q^n)                     */
EXT_MAINPROG pPOLYNOM pp_gen; /* generator polynomial g(D)                 */
EXT_MAINPROG pPOLYNOM pp_prim;/* primitive polynomial                      */
EXT_MAINPROG int      N;      /* block length coder output                 */
EXT_MAINPROG int      K;      /* block length coder input                  */
EXT_MAINPROG int      t;      /* # correctable symbols errors per block    */
EXT_MAINPROG int      bps;    /* bits per symbol                           */
EXT_MAINPROG char     InFile[STRING_LENGTH];
EXT_MAINPROG char     OutFile[STRING_LENGTH];

EXT_STREAM   long     length; /* length of input file (in bits)            */

#endif



/* 
** END OF GALOIS.H
*/ 
