/*
**    FILE    : mainprog.c 
**    PURPOSE : This file is the main-module for making 'galois.exe', a
**              simulator for error-correction codes.
**    AUTHOR  : Bart De Canne, Barco NV.
**    COMPILER: BC++ 3.1 compiler
**    DEPENDENCIES:
**              Include-files: galois.h
**              Link: stream.obj,galois.obj,generate.obj,
**              encode.obj,decode.obj
**    DATE    : 01-09-1994
**    VERSION : 1.0
*/


#define _MAINPROG
#include "galois.h"

#define BAD_INPUT { fflush(stdin);printf("\nBad input read!");break;}


static void ScrambleEncodedData(int,int);
static void ShowOpeningScreen(void);
static void ShowMainMenu(void);
static void ShowInfo(char);
void main(void);

static bool GFSymbolLoaded=FALSE,GFLocatorLoaded=FALSE,BCHConstructed=FALSE;

static void ScrambleEncodedData(int blck_no,int symbol_no) {

   int symbol_cnt,symbol,n;
   long cnt=0;

   OpenInputStream(InFile);
   OpenOutputStream(OutFile);
   START_PACIFIER;
   length /= (long)N*bps;
   do{
      PACIFIER(100*cnt/length);
      n=N;
      symbol_cnt=symbol_no;
      while(--n>=0) {
         symbol=GetBits(bps);
         if(blck_no && symbol_cnt)
            symbol=~symbol;         /* symbol error = all bits inverted */
         PutBits(symbol,bps);
         if(symbol_cnt) symbol_cnt--;
         if(EOFReached()) break;
         }
      if(blck_no) blck_no--;
      cnt++;
      }
   while(!EOFReached());
   PutBits(0,7);
   CloseInputStream();
   CloseOutputStream();
   STOP_PACIFIER;
   }

static void ShowOpeningScreen(void) {

   clrscr();
   printf(
      "\n                 "
      "\n                                                   "
      "\n                     Galois Field Arithmetic       "
      "\n                              And                  "
      "\n                       The Generation of           "
      "\n                       Error Correcting            "
      "\n                  Bose-Chaudhuri-Hocquenghem (BCH) "
      "\n                   and Reed-Solomon (RS) codes     "
      "\n                                                   "
      "\n                 "
      "\n                  "
      "\n"
      "\nAUTHOR : Bart De Canne"
      "\nCONTACT: e-mail: bdc@barco.be"
      "\n         fax   : +32-56-233 334"
      "\n         tel   : +32-56-233 211 (8-17h. W-European time)"
      "\nDATE   : 01-03-1994"
      "\n"
      "\nGalois version 1.0, Copyright (c) 1994 Bart De Canne comes with"
      "\nABSOLUTELY NO WARRANTY. This is free software and you are welcome"
      "\nto redistribute it under certain conditions."
      "\nSee the appropriate menu item for details.");
   while(!getch());
   }

static void ShowCopyright(void) {

   clrscr();
   printf(
"\n      NO WARRANTY"
"\n      -----------"
"\nBECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY FOR"
"\nTHE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. THE PROGRAM IS"
"\nPROVIDED 'AS IS' WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR "
"\nIMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF"
"\nMERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE."
"\nIN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITTEN"
"\nWILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR"
"\nREDISTRIBUTE THE PROGRAM AS PERMITTED BY THE 'TERMS OF DISTRIBUTION"
"\nAND MODIFICATION', BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL,"
"\nSPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR"
"\nINABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF DATA"
"\nOR DATA BEING RENDERED INACURRATE OR LOSSES SUSTAINED BY YOU OR THIRD "
"\nPARTIES), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE "
"\nPOSSIBILITY OF SUCH DAMAGES."
"\n");
   while(!getch());
   clrscr();
   printf(
  "      TERMS OF DISTRIBUTION AND MODIFICATION"
"\n      --------------------------------------"
"\nYOU MAY COPY AND DISTRIBUTE COPIES OF THIS PROGRAM'S SOURCE OR EXECUTABLE"
"\nCODE AS YOU RECEIVE IT, PROVIDED THAT YOU APPROPRIATELY PUBLISH ON EACH"
"\nCOPY AN APPROPRIATE COPYRIGHT NOTICE AND DISCLAIMER OF WARRANTY. "
"\nKEEP INTACT ALL THE NOTICES THAT REFER TO THIS LICENSE AND TO THE ABSENCE"
"\nOF ANY WARRANTY. YOU MAY MODIFY YOUR COPY OR COPIES OF THE PROGRAM OR ANY"
"\nPORTION OF IT, THUS FORMING A WORK BASED ON THIS PROGRAM, AND COPY AND"
"\nDISTRIBUTE SUCH MODIFICATIONS OR WORK, PROVIDED THAT YOU ALSO MEET ALL OF"
"\nTHESE CONDITIONS:"
"\n"
"\na. YOU MUST CAUSE THE MODIFIED FILES TO CARRY PROMINENT NOTICES STATING"
"\nTHAT YOU CHANGED THE FILES AND THE DATE OF ANY CHANGE, LEAVING PREVIOUS"
"\nNOTICES UNCHANGED, INCLUDING ANY REFERENCES TO CONTRIBUTING AUTHORS."
"\nb. YOU MUST CAUSE ANY WORK THAT YOU DISTRIBUTE OR PUBLISH, THAT IN WHOLE"
"\nOR IN PART CONTAINS OR IS DERIVED FROM THIS PROGRAM OR ANY PART THEREOF,"
"\nTO BE LICENSED AS A WHOLE AT NO CHARGE TO ALL THIRD PARTIES UNDER THE"
"\nTERMS OF THIS LICENSE."
"\nc. YOU SHOULD INFORM THE ORIGINAL AUTHOR OF THE PROGRAM OF ANY CHANGES"
"\nMADE BY SENDING AN E-MAIL MESSAGE TO bdc@barco.be OR FAX MESSAGE TO"
"\n+32-56-233 334."
"\n\nEND OF TERMS AND CONDITIONS.");
while(!getch());
   }
 
static void ShowMainMenu(void) {

   clrscr();
   printf(
      "GALOIS FIELD ARITHMETIC & ERROR CONTROL CODING\n"
      "==============================================\n"
      "\n"
      "                   Menu\n"
      "0. Exit\n"
      "1. Load locator field from disk -- use it as new symbol field\n"
      "2. Load symbol and locator field from disk\n"
      "3. Construct new symbol and locator field\n");
   if(GFSymbolLoaded && GFLocatorLoaded) {
      printf(  "4. Generate BCH or Reed-Solomon Code\n");
      printf(  "5. Save symbol and locator field to disk\n");
      }
   if(BCHConstructed) {
      printf(  "6. Encode\n");
      printf(  "7. Decode\n");
      printf(  "8. Scramble encoded data\n");
      }
   printf("\n9. Copyright notice/Warranty disclaimer\n");
   printf("\n\nEnter 'h' for help on using this program."
      "\nEnter 'i' for additional information on error-control coding.\n\n");
   }

static void ShowInfo(char c) {

   FILE *fp;
   char line[80];

   BEGIN("ShowInfo");
   clrscr();
   if(NULL==(fp=fopen((c=='h')?"manual.txt":"bckgrnd.txt","rt"))) ERROR;
   while(!feof(fp)) {
      clrscr();
      while(!feof(fp)) {
         fgets(line,80,fp);
         if(line[0]=='$') break;
         fputs(line,stdout);
         }
      while(!getch());
      }
   fclose(fp);
   }

/* ########################  MAIN PROGRAM  ############################### */


void main(void) {

   int cnt,i,j,k,p_s,q_s,p_l,q_l;
   char c,File[STRING_LENGTH];

   ShowOpeningScreen();
   while(1) {
   do{
      ShowMainMenu();
      printf("\nEnter choice  : ");
      c=getche();
      if(c=='h'||c=='i') {
         ShowInfo(c);
         continue;
         }
      }while (c<'0'||c>'9');
   c-=48;
   if(c<4) {
      pg_s=FreeGalois(pg_s);pg_l=FreeGalois(pg_l);
      if(po_l && po_s) {
         if(po_s->Add != po_l->Add) {
            po_s=FreeOpTable(po_s);
            po_l=FreeOpTable(po_l);
            }
         else {
            po_s=FreeOpTable(po_s);
            free(po_l);
            po_l=NULL;
            }
         }
      pp_gen=FreePolynom(pp_gen);pp_prim=FreePolynom(pp_prim);
      GFSymbolLoaded=GFLocatorLoaded=BCHConstructed=FALSE;
      }
   ResetBuffers();
   clrscr();

   switch(c) {

      case 0:                                                     /* EXIT  */
         exit(0);

      case 1:                             /* LOAD SYMBOL FIELD FROM DISK   */
         printf("LOAD LOCATOR FIELD - USE IT AS NEW SYMBOL FIELD\n");
         printf("-----------------------------------------------\n");
         AskFileName(File,NULL);
         ReadGFHeaderFromDisk(File,&p_s,&q_s,&p_l,&q_l);
         pg_s=AllocGalois(pg_s,p_l,q_l);
         po_s=AllocOpTable(po_s,p_l,q_l);
         ReadGFFromDisk(NULL,NULL,pg_s,po_s,File);
         GFSymbolLoaded=TRUE;
         goto CONSTRUCT_LOCATOR_FIELD;

      case 2:                    /* LOAD SYMBOL & LOCATOR FIELD FROM DISK  */
         printf("LOAD SYMBOL AND LOCATOR FIELD\n");
         printf("-----------------------------\n");
         AskFileName(File,NULL);
         ReadGFHeaderFromDisk(File,&p_s,&q_s,&p_l,&q_l);
         pg_s=AllocGalois(pg_s,p_s,q_s);
         po_s=AllocOpTable(po_s,p_s,q_s);
         pg_l=AllocGalois(pg_l,p_l,q_l);
         po_l=AllocOpTable(po_l,p_l,q_l);
         ReadGFFromDisk(pg_s,po_s,pg_l,po_l,File);
         GFSymbolLoaded=GFLocatorLoaded=TRUE;
         break;
         
      case 3:                          /* CONSTRUCT SYMBOL & LOCATOR FIELD */

/* 
** 1st step: Define symbol field GF(p) and operations * & + in it
** 2nd step: Search for a primitive polynomial, generating
**           all elements of GF(p^q)
** 3rd step: Search for the minimal polynomial of each GF-element
*/ 
   /* 1.*/
         printf("CONSTRUCTION OF GF(p^q)\n"
                "-----------------------\n"
                "\n"
                "\nEnter p : ");
         p_s=getche();
         if(p_s<'2'||p_s>'9') {
            printf("\np should be a prime integer in the range 2..9");
            break;
            }
         p_s-=48;
         q_s=1;
         if(!IsPrime(p_s)) {
            printf("\nSorry, p should be prime!");
            break;
            }
         pg_s=AllocGalois(pg_s,p_s,q_s);
         po_s=AllocOpTable(po_s,p_s,q_s);
                        /* elements of symbol field always denoted 0..p-1  */
                        /* and operations defined modulo p                 */
         for(i=0;i<p_s;i++) 
            pg_s->pe[i].Alpha[0]=i;
         printf("\nConstructing GF(%d^%d) table of operations...",
            pg_s->Base,pg_s->Exp);
         po_s=ConstructOpTable(po_s,pg_s,NULL);
         printf("OK");
         GFSymbolLoaded=TRUE;

CONSTRUCT_LOCATOR_FIELD:
   /* 2.*/

         p_l=(int)pow((double)pg_s->Base,(double)pg_s->Exp);
         printf("\nEnter q : ");
         q_l=getche();
         if(q_l<'1'||q_l>'8') {
            printf("\nq should be an integer in the range 1..8!");
            break;
            }
         q_l-=48;
         if((int)pow((double)p_l,(double)q_l) > 256) {
            printf("\nField-size too large: p^q should not exceed 256!");
            break;
            }
         pg_l=AllocGalois(pg_l,p_l,q_l);
         if(q_l>1) 
            po_l=AllocOpTable(po_l,p_l,q_l);
         else {               /* no need to allocate new storage space     */
                              /* since symbol and locator fields are       */
                              /* the same                                  */
            po_l=(pOPTABLE)malloc(sizeof(OPTABLE));
            po_l->MaxNo = pg_l->MaxNo;
            po_l->Base=p_l;
            po_l->Exp=q_l;
            po_l->Add=po_s->Add;
            po_l->Mult=po_s->Mult;
            po_l->InvAdd=po_s->InvAdd;
            po_l->InvMult=po_s->InvMult;
            }
         pp_prim=AllocPolynom(pp_prim,pg_l->Base,pg_l->Exp);
         printf("\nSearching for GF(%d^%d) primitive polynomial...",
            pg_l->Base,pg_l->Exp);
         FindPrimitivePolynomial();
         printf("OK\nPrimitive polynomial in use :\n");
         PrintPoly(pp_prim);
         printf("\nConstructing GF(%d^%d) table of operations...",
            pg_l->Base,pg_l->Exp);
         po_l=ConstructOpTable(po_l,pg_l,po_s);
                                       /* and po_l defines operations   */
                                       /* of GF(p^q)...                 */
   /* 3.*/
         printf("\nSearching for GF(%d^%d) minimal polynomials...",
            pg_l->Base,pg_l->Exp);
         FindMinimalPolynomials();
                                       /* and pg_l contains pointers to */
                                       /* min.pol. of each element.     */
         printf("OK\nGF constructed.");
         GFLocatorLoaded=TRUE;
         break;

      case 4: 
/*
** BCH-code generated by using the array of GF-elements and their
** minimal polynomials
*/

         printf("GENERATION OF BCH / REED-SOLOMON CODE\n");
         printf("-------------------------------------");
         if(!GFSymbolLoaded || !GFLocatorLoaded) break;
         N=pg_l->MaxNo-1;
         if(pg_l->Exp==1)
            printf("\nCode will be REED-SOLOMON.");
         printf(
            "\n"
            "\nblocklength N=p^q-1 (%d)"
            "\ncan correct up to t-symbol errors (t<=N/2 (%d))",N,N>>1);

         printf("\nEnter t (t<=%d): ",N>>1);
         if(!scanf("%d",&t)) BAD_INPUT;
         if(t<1 || t>(N>>1)) BAD_INPUT;  
         pp_gen=AllocPolynom(pp_gen,p_l,2*t*q_l);

         printf("\nGenerating %d symbol-error correcting BCH Code",t);
         GenerateBCH();
         K=N-pp_gen->Deg;
         bps=(int)(log10((double)pp_gen->Base)/log10(2.0));

         printf("\nGenerator polynomial :\n");
         PrintPoly(pp_gen);

         printf("\nN=%d,N-K=%d of %d-ary symbols",N,N-K,pg_l->Base);
         BCHConstructed=TRUE;
         break;

      case 5: 
         if(!GFSymbolLoaded || !GFLocatorLoaded) break;
         AskFileName(File,NULL);
         WriteGFToDisk(File);
         break;

      case 6:
      case 7:

         if(!BCHConstructed) break;
         printf("\nEncoding / decoding of bit-stream"
                "\n---------------------------------");

         if(p_l%2) {
            printf("\nCan only encode/decode in GF(p^q) where "
                   "p is a power of 2...");
            break;
            }

         printf("\nFile to encode / decode");
         AskFileName(InFile,(c==6)?("test.in"):("test.cod"));
         printf("\nOutput file");
         AskFileName(OutFile,(c==6)?("test.cod"):("test.rec"));

         if(c==6) {                                               /* ENCODE */
            printf("\nEncoding (N=%d,K=%d) with %d-bits per symbol",N,K,bps);
            EncodeData();
            }
         else {                                                   /* DECODE */
            printf("\nDecoding (N=%d,K=%d) with %d-bits per symbol",N,K,bps);
            DecodeData();
            PrintDecodeStatistics();
            }
         break;

      case 8: {
         int err,no;

         if(!BCHConstructed) break;
         printf("\nScrambling of encoded bitstream"
                "\n-------------------------------");
         if(p_l%2) {
            printf("\nCan only scramble in GF(p^q) where "
                   "p is a power of 2...");
            break;
            }
         printf("\nFile to scramble. ");
         AskFileName(InFile,"test.cod");
         printf("\nOutput file ");
         AskFileName(OutFile,"test.scr");
         printf("\n# blocks to scramble ? ");
         if(!scanf("%d",&no)) BAD_INPUT;
         printf("\n# symbol errors per block ? ");
         if(!scanf("%d",&err)) BAD_INPUT;
         printf("\n\nScrambling ..."); 
         ScrambleEncodedData(no,err);
         printf("OK.\n");
         break;}
         
      case 9: 
         ShowCopyright();
         break;

      default:
         break;
      }
   printf("\nPress key to continue...");
   while(!getch()) ;
   }
   }


/* 
** END OF MAINPROG.C
*/ 
